import bisect

from train_route.path import Path

class Schedule:
    def __init__(self, schedule_id, path:Path, progress_points:list, timestamps:list) -> None:
        self.id = schedule_id
        self.path = path
        self.progress_points = progress_points
        self.timestamps = timestamps

        self.start_time = timestamps[0]
        self.end_time = timestamps[-1]

    def progress(self, time):
        if time < self.start_time or time > self.end_time:
            raise BaseException(
                f'Time outside schedule: {time} not in [{self.start_time}, {self.end_time}]'
            )
        if time == self.start_time:
            return self.progress_points[0]
        index = bisect.bisect_left(self.timestamps, time)
        if index == 0 or index == len(self.timestamps):
            raise BaseException(
                f'Debug: Bisect should return valid index from 1 to len - 1'
            )
        t0 = self.timestamps[index - 1]
        t1 = self.timestamps[index]
        delta_t = (time - t0) / (t1 - t0)
        if delta_t < 0 or delta_t > 1:
            raise BaseException(
                f'Debug: delta ({delta_t}) should be between 0 and 1: t0 = {t0}, t1 = {t1}'
            )
        p0 = self.progress_points[index - 1]
        p1 = self.progress_points[index]
        return p0 + delta_t * (p1 - p0)
    
    def position(self, time):
        progress = self.progress(time)
        return self.path.position(progress)
    
    def chain_state(self, time, chain_relative_progress):
        progress = self.progress(time)
        chain_progress = [progress + p for p in chain_relative_progress]
        return self.path.chain_state(chain_progress)
    
    def is_valid(self, time):
        "test if time is valid"
        return time >= self.start_time and time <= self.end_time



from .train import Train
from .utils import is_inside

class TrainRender:
    def __init__(self, train:Train) -> None:
        self.train = train
        self.current_position = None
        self.chain_state = None
        "position on screen"
        self.status:str = 'new'
    
    def set_time(self, time, window):
        if not self.train.is_active(time):
            self.status = 'inactive'
            return
            
        position = self.train.position(time)
        if self.current_position is None:
            self.status = 'active'
            self.current_position = position
            if self.train.length > 0.01 * window['radius']:
                self.chain_state = self.train.chain_state(time)
            else:
                self.chain_state = None
            return
        if position == self.current_position:
            self.status = 'sleep'
            return
        if not (is_inside(self.current_position, window) or is_inside(position, window)):
            self.status = 'sleep'
            return

        self.status = 'active'
        self.current_position = position
        if self.train.length > 0.01 * window['radius']:
            self.chain_state = self.train.chain_state(time)
        else:
            self.chain_state = None

    def to_dict(self):
        result = {
            'train_id': self.train.id,
            'status': self.status,
        }
        if self.status == 'active':
            result['position'] = self.current_position
            result['chain_state'] = self.chain_state
        return result



from train_route.schedule import Schedule

class Train():
    def __init__(
            self,
            train_id,
            schedule:Schedule,
            chain_relative_progress:list,
            start
    ) -> None:
        self.id = train_id
        self.schedule = schedule
        self.chain_relative_progress = chain_relative_progress
        self.start = start
        self.length = abs(chain_relative_progress[0] - chain_relative_progress[-1])

    def position(self, time):
        "position of head of chain"
        relative_time = time - self.start
        return self.schedule.position(relative_time)
    
    def chain_state(self, time):
        "state of chain"
        relative_time = time - self.start
        return self.schedule.chain_state(relative_time, self.chain_relative_progress)
    
    def is_active(self, time):
        "test if train has position"
        relative_time = time - self.start
        if not self.schedule.is_valid(relative_time):
            return False
        progress = self.schedule.progress(relative_time)
        if progress < self.length:
            return False
        return True
    



from .train import Train
from .train_render import TrainRender
from typing import List
from .utils import spherical_distance

class TrainSystem:
    def __init__(
            self, 
            trains:List[Train],
            center,
            radius,
            distance_function = spherical_distance
            ) -> None:
        self.train_renders = [TrainRender(train) for train in trains]
        self.center = center
        self.radius = radius
        self.distance_function = distance_function
    
    def window(self):
        return {
            'center': self.center,
            'radius': self.radius,
            'distance': self.distance_function,
        }

    def frame_data(self, time):
        for render in self.train_renders:
            render.set_time(time, self.window())
        return [
            render.to_dict() for render in self.train_renders
        ]



import bisect

from .utils import cumulative_distance, interpolate

class Segment:
    "Segment is used for inner paths with missing time points"
    def __init__(self, 
                 segment_id, 
                 source, 
                 target, 
                 latitudes, 
                 longitudes, 
                 cumulative_distances = None
                 ) -> None:
        self.id = segment_id
        self.source = source
        self.target = target
        self.latitudes = latitudes
        self.longitudes = longitudes
        if cumulative_distances:
            self.cumulative_distances = cumulative_distances
        else:
            self.cumulative_distances = cumulative_distance(latitudes, longitudes)
        self.total_distance = self.cumulative_distances[-1]
    
    def position(self, progress):
        if progress < 0 or progress > self.total_distance:
            if abs(progress - self.total_distance) < 1e-8:
                return [self.latitudes[-1], self.longitudes[-1]]
            raise BaseException(
                f'Position outside segment: {progress} not in [{0}, {self.total_distance}]'
            )
        if progress == 0:
            return [self.latitudes[0], self.longitudes[0]]
        index = bisect.bisect_left(self.cumulative_distances, progress)
        if index == 0 or index == len(self.cumulative_distances):
            raise BaseException(
                f'Debug: Bisect should return valid index from 1 to len - 1'
            )
        d0 = self.cumulative_distances[index - 1]
        d1 = self.cumulative_distances[index]
        delta = (progress - d0) / (d1 - d0)
        if delta < 0 or delta > 1:
            raise BaseException(
                f'Debug: delta ({delta}) should be between 0 and 1: d0 = {d0}, d1 = {d1}'
            )
        latitude = interpolate(self.latitudes, index, delta)
        longitude = interpolate(self.longitudes, index, delta)
        return [latitude, longitude]
    
    def position_from_ratio(self, progress_ratio):
        if progress_ratio < 0 or progress_ratio > 1:
            raise BaseException(
                f'Debug: progress ratio ({progress_ratio}) should be between 0 and 1'
            )
        progress = self.total_distance * progress_ratio
        return self.position(progress)

    def to_dict(self):
        return {
            'segment_id': self.segment_id,
            'source': self.source,
            'target': self.target,
            'latitudes': self.latitudes,
            'longitudes': self.longitudes,
            'cumulative_distances': list(self.cumulative_distances),
        }
    
    def to_ref(self):
        return self.segment_id


from typing import List
import bisect

from .segment import Segment
from .utils import cumsum, state

class Path:
    def __init__(self, path_id, segments:List[Segment]) -> None:
        self.id = path_id
        self.segments = segments
        
        self.cumulative_distances = cumsum(segment.total_distance for segment in segments)
        self.total_distance = self.cumulative_distances[-1]

    def position(self, progress):
        if progress < 0 or progress > self.total_distance:
            raise BaseException(
                f'Position outside segment: {progress} not in [{0}, {self.total_distance}]'
            )
        index = bisect.bisect_left(self.cumulative_distances, progress)
        if index == len(self.cumulative_distances):
            raise BaseException(
                f'Debug: Bisect should return valid index from 0 to len - 1'
            )
        if index == 0:
            segment_progress_start = 0.
        else:
            segment_progress_start = self.cumulative_distances[index - 1]
        segment_progress = progress - segment_progress_start
        return self.segments[index].position(segment_progress)
    
    def chain_state(self, chain_progress):
        positions = [self.position(progress) for progress in chain_progress]
        positions_start = positions[:-1]
        positions_end = positions[1:]
        return [state(start, end) for start, end in zip(positions_start, positions_end)]



import math
import pyproj

geod = pyproj.Geod(ellps='WGS84')

def plane_distance(lats, lons):
    x0, x1 = lats
    y0, y1 = lons
    return math.sqrt((x1 - x0)**2 + (y1 - y0)**2)

def spherical_distance(lats, lons):
    return geod.line_lengths(lons, lats)[0]

def spherical_rotation(lats, lons):
    x0, x1 = lats
    y0, y1 = lons
    fwd_azimuth, back_azimuth, distance = geod.inv(y0, x0, y1, x1)
    return fwd_azimuth

def cumulative_distance(latitudes, longitudes, *, distance=plane_distance):
    distances = [0.]
    for i in range(1, len(latitudes)):
        lats = latitudes[i - 1:i + 1]
        lons = longitudes[i - 1:i + 1]
        local_distance = distance(lats, lons)
        distances.append(local_distance)
    return cumsum(distances)

def interpolate(xs, index, delta):
    x0 = xs[index - 1]
    x1 = xs[index]
    return x0 + delta * (x1 - x0)

def cumsum(xs:list):
    cs = list(xs)
    for i in range(1, len(cs)):
        cs[i] += cs[i - 1]
    return cs

def state(start, end):
    lat0, lon0 = start
    lat1, lon1 = end
    return [
        (lat0 + lat1) / 2,
        (lon0 + lon1) / 2,
        spherical_rotation((lat0, lat1), (lon0, lon1))
    ]

def is_inside(position, window):
    lat, lon = position
    d = window['distance']([window['center']['lat'], lat], [window['center']['lon'], lon])

    if d <= window['radius']:
        return True
    else:
        return False